use data::tables;
use proto::*;
use thiserror::Error;

use crate::logic::{math::Vector3f, time::MainCityTime, ESceneType};

use super::NapGameMode;

pub struct FrontendGame {
    section_id: u32,
    frontend_avatar_id: i32,
    camera_x: u32,
    camera_y: u32,
    born_pos: String,
    main_city_time: MainCityTime,
    avatar_pos: Vector3f,
    avatar_rot: Vector3f,
}

#[derive(Error, Debug)]
pub enum FrontendGameError {
    #[error("section id is invalid ({0})")]
    InvalidSection(u32),
}

impl FrontendGame {
    pub fn new(
        section_id: u32,
        avatar_id: i32,
        main_city_time: MainCityTime,
        avatar_pos: Vector3f,
        avatar_rot: Vector3f,
    ) -> Result<Self, FrontendGameError> {
        let instance = Self {
            section_id,
            main_city_time,
            frontend_avatar_id: avatar_id,
            camera_x: 0xFFFFFFFF,
            camera_y: 0xFFFFFFFF,
            born_pos: Self::get_default_stage_entry_name(section_id)
                .ok_or(FrontendGameError::InvalidSection(section_id))?,
            avatar_pos,
            avatar_rot,
        };

        tracing::info!("creating new frontend game (section={section_id}, avatar={avatar_id})");
        Ok(instance)
    }

    fn get_default_stage_entry_name(section_id: u32) -> Option<String> {
        tables::section_config_template_tb::iter()
            .find(|tmpl| tmpl.section_id == section_id)
            .map(|tmpl| tmpl.primary_entry_name.clone())
    }
}

impl NapGameMode for FrontendGame {
    fn scene_info(&self) -> Option<SceneInfo> {
        Some(SceneInfo {
            scene_type: self.scene_type() as u32,
            hall_scene_info: Some(HallSceneInfo {
                section_id: self.section_id,
                frontend_avatar_id: self.frontend_avatar_id as u32,
                camera_x: self.camera_x,
                camera_y: self.camera_y,
                born_pos: self
                    .avatar_pos
                    .is_zero()
                    .then_some(self.born_pos.clone())
                    .unwrap_or_default(),
                day_of_week: self.main_city_time.day_of_week.value(),
                time_of_day: self.main_city_time.time_of_day.value(),
                position: (!self.avatar_pos.is_zero()).then_some(Transform {
                    position: self.avatar_pos.to_vec(),
                    rotation: self.avatar_rot.to_vec(),
                }),
                ..Default::default()
            }),
            ..Default::default()
        })
    }

    fn dungeon_info(&self) -> Option<DungeonInfo> {
        None
    }

    fn scene_type(&self) -> ESceneType {
        ESceneType::Hall
    }
}
